"""Constants for the WeLock integration."""

OAUTH2_AUTHORIZE = "https://iot.we-lock.com/auth/login"
OAUTH2_TOKEN = "https://iot.we-lock.com/auth/Authorize"
OAUTH2_CLIENT_ID = ""
MQ_CONFIG_URL = "url"
MQ_CONFIG_CLIENT_ID = "clientId"
MQ_CONFIG_USER_NAME = "username"
MQ_CONFIG_PASSWORD = "password"
MQ_CONFIG_TOPIC = "topic"
MQ_CONFIG_EXP = "expireTime"

SENSOR_BATTERY_KEY = "Battery"
SENSOR_RECORD_KEY = "Record"
SWITCH_WIFIBOX_KEY = "RemoteControl"
BUTTON_WIFIBOX_KEY = "RemoteControl"
LOCK_WELOCK_KEY = "LockOperation"
BINARYSENSORS_DOOR_KEY = "status"



"""Manager res and data."""

import datetime
from typing import Any

from .client import WeLockApi
from .const import BINARYSENSORS_DOOR_KEY, SENSOR_BATTERY_KEY, SENSOR_RECORD_KEY
from .models import (
    OPENTYPE,
    WeLockData,
    WeLockDevice,
    WeLockException,
    WeLockMessageListener,
    logger,
)
from .wemq import WeMqttClient


class Resmanager:
    """Manager res and data."""

    def __init__(self, api: WeLockApi) -> None:
        """Init."""
        self.api = api
        self.mqclient = None
        self.device_map: dict[str, WeLockDevice] = {}
        self.device_listeners = set()

    async def initWeMq(self):
        """Init mq and start mq."""
        if self.mqclient is not None:
            await self.mqclient.disconnect()
            self.mqclient = None
        if self.api is not None:
            mq = WeMqttClient(self.api)
            await mq.connect(self.on_message)
            self.mqclient = mq

    def on_message(self, msg_data: dict):
        """Receive message."""
        logger.debug(f"mq receive-> {msg_data}")
        self._process_user_data(msg_data)

    def __convert_to_welockdata(self, status: list, device_id: str) -> dict:
        values = {}
        dev = self.device_map.get(device_id, None)
        if dev:
            values = dev.status or {}

        for item in status:
            we_data = WeLockData()
            types = item.get("type", "").strip()
            we_data.type = types
            if types in ("Unlock", "RelayAction"):
                we_data.type = SENSOR_RECORD_KEY
            if types == "Battery":
                we_data.type = SENSOR_BATTERY_KEY
            we_data.unlock_type = item.get("unlockType", 0)
            if we_data.unlock_type == 31:
                we_data.type = BINARYSENSORS_DOOR_KEY
            if we_data.type:
                we_data.unlock_result = item.get("unlockResult", 0)
                we_data.result_desc = item.get("resultDescription", None)
                we_data.user_id = f"{item.get('userId', '')}"
                if types == "Unlock":
                    we_data.unlock_time = item.get("catchTime", item.get("time", 0))
                    we_data.unlock_type = item.get("unlockType", 17)
                    we_data.user_remark = item.get("uRemark", None)
                elif types == "RelayAction":
                    we_data.unlock_time = item.get("time", item.get("catchTime", 0))
                    we_data.user_id = item.get("userId", device_id)
                    we_data.user_remark = item.get("uRemark", None)
                we_data.value = self.__getDataValue(we_data)
                values[we_data.type] = we_data.value
        return values

    def __getDataValue(self, we_data: WeLockData) -> Any:
        """Show detail info."""
        value = None
        try:
            if we_data.type == SENSOR_RECORD_KEY:
                value = f"unlocked by {OPENTYPE.get(we_data.unlock_type, '')}({we_data.user_remark if we_data.user_remark is not None else we_data.user_id}) at {self.__dateFormat(we_data.unlock_time)}"
            elif we_data.type == BINARYSENSORS_DOOR_KEY:
                value = we_data.unlock_result == 1
            elif we_data.type == SENSOR_BATTERY_KEY:
                value = int(we_data.result_desc, 16)
        except AttributeError:
            pass
        except WeLockException:
            pass
        return value

    def __dateFormat(self, timestamp_ms) -> str:
        """Datetime format."""
        timestamp = timestamp_ms / 1000.0
        dt_object = datetime.datetime.fromtimestamp(timestamp)
        return dt_object.strftime("%Y-%m-%d %H:%M:%S")

    def _process_user_data(self, data: dict) -> None:
        """Parse the data."""
        try:
            device_id = data.get("number", "")
            status = data.get("status", [])
            device = self.device_map.get(device_id, None)
            if device:
                if isinstance(status, list):
                    values = self.__convert_to_welockdata(status, device_id)
                    device.status = values
            for listener in self.device_listeners:
                listener.update_device(device)
        except KeyError as ke:
            logger.error("_process_user_data1 = %s", ke)
        except TypeError as er:
            logger.error("_process_user_data2 = %s", er)

    def add_device_listener(self, listener: WeLockMessageListener):
        """Add device listener."""
        self.device_listeners.add(listener)

    async def update_device_list(self):
        """Return the locks."""
        self.device_map.clear()
        if device_list := await self.api.get_locks():
            for device in device_list:
                self.device_map[device.device_id] = device

    async def unlock(self, device: WeLockDevice) -> bool:
        """."""
        await self.api.unlock(device)
        return True

    async def remote_control(self, device: WeLockDevice) -> bool:
        """."""
        await self.api.remoteControl(device)
        return True



"""WeLock api for iot ."""

import asyncio
from collections.abc import Mapping
from typing import Any, cast
from urllib.parse import urljoin

from aiohttp import ClientError, ClientResponse, ClientSession

from .auth_token_manager import OauthTokenManager
from .models import (
    DeviceType,
    RequestResult,
    RequestResultStatus,
    WeLockDevice,
    WeLockRequestException,
    logger,
)

RQ_LOCK = asyncio.Lock()


class WeLockApi:
    """Api for welock."""

    BASE = "https://iot.we-lock.com/"

    def __init__(self, oauth_mgr: OauthTokenManager) -> None:
        """Initialize WeLockApi auth."""
        self._oauth_mgr = oauth_mgr

    @property
    async def _async_get_access_token(self) -> str:
        """Return a valid access token."""
        return await self._oauth_mgr.check_and_refresh_token()

    @property
    def _web_session(self) -> ClientSession:
        """Return the client session."""
        return self._oauth_mgr.client_session()

    async def _parse_resp(self, resp: ClientResponse) -> Mapping[str, Any]:
        """Parse the response content."""
        body = ""
        status_code = resp.status
        if status_code == 200:
            body = await resp.json()
        else:
            try:
                resp.raise_for_status()
            except ClientError as client_err:
                body = await resp.text()
                logger.debug("API returned: %s", body)
                raise WeLockRequestException("-1", "request failed!") from client_err
            except WeLockRequestException:
                raise

        return cast(dict, await resp.json())

    def _result_resp(self, resp: Mapping[str, Any]) -> RequestResult:
        """Convert response result ."""
        rq = RequestResult(None, None, None)
        rq.code = -1
        if resp:
            rq.code = resp.get("code", -1)
            rq.data = resp.get("data", None)
            rq.msg = resp.get("msg", None)

        return rq

    async def request(
        self, method: str, path: str, auth_required: bool = True, **kwargs: Any
    ) -> ClientResponse:
        """Add headers."""
        headers = kwargs.pop(
            "headers", {"Content-Type": "application/x-www-form-urlencoded"}
        )
        params = kwargs.pop("params", None)
        data = kwargs.pop("data", None)
        url = urljoin(self.BASE, path)

        extra_headers = kwargs.pop("extra_headers", None)
        extra_params = kwargs.pop("extra_params", None)
        extra_data = kwargs.pop("extra_data", None)
        if auth_required:
            access_token = await self._async_get_access_token
            headers["Authorization"] = f"{access_token}"

        if extra_headers:
            headers.update(extra_headers)
        if extra_params:
            params = params or {}
            params.update(extra_params)
        if extra_data:
            data = data or {}
            data.update(extra_data)
        return await self._web_session.request(
            method, url, **kwargs, headers=headers, params=params, data=data, timeout=8
        )

    async def post(self, path: str, params: Mapping[str, Any]) -> Mapping[str, Any]:
        """Post request."""
        resp = await self.request("post", path, True, params=params)
        return await self._parse_resp(resp)

    def __parse_to_list(self, data: list) -> list[WeLockDevice]:
        """Convert to welock device."""
        device_list = []
        for item in data:
            device = WeLockDevice()
            device.device_id = item.get("number", None)
            device.device_name = item.get("notes", None) or f"{device.device_id}"
            device.model_name = item.get("model", None)
            device.device_type = DeviceType.value2enum(item.get("deviceType", 1))
            gateway = item.get("gatewayModel", None)
            device.model_name2 = gateway
            if (
                gateway
                and gateway != "WIFIBOX3"
                and device.device_type == DeviceType.WIFIBOX3
            ):
                device.device_type = DeviceType.NOTSUPPORT
            bleInfos = item.get("bluetooth", None)
            if isinstance(bleInfos, list):
                if len(bleInfos) == 1:
                    device.ble_name1 = bleInfos[0].get("deviceName", None)
                    device.battery = bleInfos[0].get("power", 1)
                else:
                    for ble in bleInfos:
                        position = ble.get("position", None)
                        deviceName = ble.get("deviceName", None)
                        power = ble.get("power", 1)
                        if position in (1, 0):
                            device.ble_name1 = deviceName
                            device.battery = power

            device_list.append(device)
        return device_list

    async def get_locks(self) -> list[WeLockDevice]:
        """Enumerate all locks in the account."""
        device_list = []
        res = await self.post("Device/List", [])

        rq = self._result_resp(res)
        if rq.code == RequestResultStatus.SUCCESS.value:
            if isinstance(rq.data, list):
                device_list = self.__parse_to_list(rq.data)
        elif rq.code == RequestResultStatus.FAILED.value:
            raise WeLockRequestException(
                RequestResultStatus.FAILED.value, "get list failed"
            )

        return device_list

    async def unlock(self, device: WeLockDevice) -> bool:
        """Unlock."""
        result = False
        params = {}
        params["number"] = device.device_id
        params["deviceName"] = device.ble_name1
        res = await self.post("Device/GatewayUnlocking", params)
        rq = self._result_resp(res)
        if rq.code == RequestResultStatus.SUCCESS.value:
            result = True
        elif rq.code == RequestResultStatus.FAILED.value:
            result = False
        return result

    async def remoteControl(self, device: WeLockDevice) -> bool:
        """Wifibox remote control."""
        result = False
        params = {}
        params["number"] = device.device_id
        res = await self.post("Device/triggerSwitch", params)
        rq = self._result_resp(res)

        if rq.code == RequestResultStatus.SUCCESS.value:
            result = True
        elif rq.code == RequestResultStatus.FAILED.value:
            result = False
        return result

    async def mqttConfig(self) -> Mapping[str, Any]:
        """Get the user mqtt config."""
        config = {}
        res = await self.post("MQTT/info", [])
        rq = self._result_resp(res)
        if rq.code == RequestResultStatus.SUCCESS.value:
            if isinstance(rq.data, Mapping):
                config = rq.data

        return config



"""Welock."""

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from enum import Enum
import logging
from typing import Any

logger = logging.getLogger(__package__)
logger.setLevel(logging.DEBUG)

logging.basicConfig(level=logging.DEBUG)

default_handler = logging.StreamHandler()
default_handler.setFormatter(
    logging.Formatter("[%(asctime)s] [welock-%(module)s] %(message)s")
)

logger.addHandler(default_handler)


class DeviceType(Enum):
    """Welock device type : lock,sensor,gateway."""

    WELOCK = 1
    WIFIBOX3 = 3
    DOORS = 6
    NOTSUPPORT = 100

    @classmethod
    def value2name(cls, value):
        """."""
        for member in cls:
            if member.value == value:
                return member.name
        return None

    @classmethod
    def value2enum(cls, value):
        """."""
        for member in cls:
            if member.value == value:
                return member
        return cls.NOTSUPPORT


class RequestResultStatus(Enum):
    """."""

    SUCCESS = 0
    FAILED = -1


OPENTYPE: dict[int, str] = {
    1: "Fp",
    2: "Passcode",
    3: "Card",
    4: "BTremote",
    5: "433",
    6: "App",
    10: "Passcode",
    11: "Tem pwd",
    12: "Tem pwd",
    13: "Tem pwd",
    14: "Passcode",
    15: "FB",
    16: "NFC",
    17: "WIFIBOX",
    19: "Vacation",
    211: "Local",
}


class WeLockDevice:
    """."""

    device_id: str
    device_name: str
    ble_name1: str
    ble_name2: str
    model_name: str
    device_type: DeviceType
    status: dict[str, Any] = {}
    battery: int
    model_name2: str

    def __init__(self) -> None:
        """."""

    @property
    def model_show(self) -> str:
        """."""
        if self.model_name:
            return self.model_name
        if self.model_name2:
            return self.model_name2
        return None


class WeLockData:
    """."""

    type: str
    unlock_type: int
    user_id: str
    unlock_result: int
    result_desc: str
    value: Any
    unlock_time: int
    user_remark: str

    def __init__(self) -> None:
        """."""


class WeLockException(Exception):
    """Exception when api returns an error."""


class WeLockRequestException(WeLockException):
    """."""

    def __init__(
        self,
        code: str,
        desc: str,
    ) -> None:
        """Initialize the api error."""

        self.code = code
        self.message = desc


@dataclass
class RequestResult:
    """."""

    code: str
    data: Any
    msg: str


class WeLockMessageListener(metaclass=ABCMeta):
    """message listener."""

    @abstractmethod
    def on_message(self, msg_data: dict[str, Any]) -> None:
        """On device message receive."""



"""Mqtt."""

import asyncio
import json
import ssl
from typing import Any
from urllib.parse import urlsplit

import aiomqtt

from .client import WeLockApi
from .models import WeLockException, logger

ssl_context = ssl.create_default_context()


class WeMqttConfig:
    """Mqtt config."""

    def __init__(self, config: dict[str, Any]) -> None:
        """Init mqconfig."""
        self.url = config.get("url", "")
        self.client_id = config.get("clientId", "")
        self.username = config.get("username", "")
        self.password = config.get("password", "")
        self.keepalive = config.get("keepalive", 60)
        self.topic = config.get("topic")


class WeMqttClient:
    """Mqtt client."""

    def __init__(self, welockapi: WeLockApi) -> None:
        """."""
        super().__init__()
        self.api = welockapi
        self.client = None
        self.mq_config = None
        self._message_listener = None
        self._running = False
        self._listener_task = None

    async def connect(self, listener) -> None:
        """Connect mqtt broker."""
        self._message_listener = listener
        self._listener_task = asyncio.create_task(self._listen())

    async def _get_config(self):
        config = await self.api.mqttConfig()
        mq_config = WeMqttConfig(config)
        self.mq_config = mq_config

    async def _listen(self):
        await self._get_config()
        reconnect_interval = 30
        self._running = True
        while self._running:
            url = urlsplit(self.mq_config.url)
            _tls_context = None
            if url.scheme == "ssl":
                _tls_context = ssl_context
            try:
                async with aiomqtt.Client(
                    hostname=url.hostname,
                    port=url.port,
                    identifier=self.mq_config.client_id,
                    username=self.mq_config.username,
                    password=self.mq_config.password,
                    keepalive=self.mq_config.keepalive,
                    tls_context=_tls_context,
                ) as mqttc:
                    self.client = mqttc
                    await self.client.subscribe(self.mq_config.topic)
                    logger.info("mqtt client connected.")
                    async for message in self.client.messages:
                        self._process_message(message)
            except aiomqtt.MqttError as mqtt_err:
                logger.error(
                    "[%s] mqtt client disconnected!",
                    self.mq_config,
                    exc_info=True,
                )
                await asyncio.sleep(reconnect_interval)
                if isinstance(mqtt_err, aiomqtt.MqttCodeError):
                    if mqtt_err.rc in [3, 4, 5]:
                        logger.error(
                            "[%s] connect failed.",
                            mqtt_err.rc,
                        )
                        await self._get_config()
            except WeLockException:
                await asyncio.sleep(reconnect_interval)
                logger.error("[%s] unexcept exception:", self.mq_config, exc_info=True)

    def _process_message(self, msg: aiomqtt.message.Message) -> None:
        """Mqtt on message."""
        logger.debug(f"{msg.topic} :payload-> {msg.payload.decode('utf8')}")
        try:
            msg_dict = json.loads(msg.payload.decode("utf8"))
            self._message_listener(msg_dict)
        except json.JSONDecodeError as ex:
            logger.error("_on_message error %s", ex)

    async def disconnect(self) -> None:
        """UnRegister listener."""
        if self._listener_task is None:
            return
        self._listener_task.cancel()
        self._listener_task = None
        self._running = False
        self.client = None



"""Authorization manager."""

import abc

from aiohttp import ClientSession


class OauthTokenManager(metaclass=abc.ABCMeta):
    """API Authentication Manager."""

    def __init__(self, session: ClientSession) -> None:
        """Auth Manager."""
        self._session = session

    def client_session(self) -> ClientSession:
        """Get client session."""
        return self._session

    @abc.abstractmethod
    def access_token(self) -> str:
        """Get auth token."""

    @abc.abstractmethod
    async def check_and_refresh_token(self) -> str:
        """Check and fresh token."""



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="welock_iot",
    version="0.0.2",
    author="hzjchina",
    author_email="hzjchina@yeah.net",
    description="welock_iot", 
    long_description=long_description,
    long_description_content_type="text/markdown",   
    # url="https://github.com/",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    python_requires=">=3.6",
    install_requires=[
        "aiohttp>=3.8.1",
        "aiomqtt>=2.0.0,<3.0.0", 
    ],
    include_package_data=False,
    license='MIT'

)